# Copyright DST Group. Licensed under the MIT license.
import re
from ipaddress import IPv4Network, IPv4Address
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.logger import CybORGLogger
from csle_cyborg.shared.actions.game_actions.game_action import GameAction


class GetTrueState(GameAction, CybORGLogger):
    """Get true state of game. """

    def __init__(self, info):
        self.info = info

    def emu_execute(self, game_controller, *args, **kwargs) -> Observation:
        agent_sessions = None
        if self.info is not None:
            obs = Observation()

            for host, data in self.info.items():

                # get systeminfo
                self._log_debug(f"{host}: {data}")
                if "Sessions" in data:

                    if data["Sessions"] == 'All':
                        for agent_name, agent_sessions in game_controller._agents.items():

                            for session in agent_sessions.get_sessions():

                                hostid = session.pop('hostid')
                                try:
                                    hostid = game_controller.network.ip_hostname_map[IPv4Address(hostid)]
                                except:
                                    pass
                                if hostid == host:
                                    obs.add_session_info(hostid=hostid, **session)
                    else:
                        agent_name = data["Sessions"]
                        if agent_sessions is None:
                            agent_sessions = game_controller._agents[agent_name].get_sessions()

                        for session in agent_sessions:

                            if session['hostid'] == host:
                                obs.add_session_info(**session)
                if "System info" in data:
                    # game_controller._log_debug(f'Getting System info for {host}')
                    result = game_controller.execute_ssh_command('systeminfo /FO csv /NH', host)
                    if result == '' or 'command not found' in result or 'Invalid command' in result:
                        # likely fail due to wrong OS
                        result = game_controller.execute_ssh_command(
                            'uname -s && uname -m && lsb_release -i && lsb_release -r && hostname', host)
                        result = result.split('\n')
                        # self._log_debug(result)
                        obs.add_system_info(hostid=host, hostname=result[4], architecture=result[1], os_type=result[0],
                                            os_distribution=result[2].split(':\t')[1],
                                            os_version=result[3].split(':\t')[1])
                    else:
                        result = result.split(',')
                        obs.add_system_info(hostid=host, OSType=result[1], OSDistribution=result[1],
                                            OSVersion=result[2], architecture=result[15].split('-')[0][1:])
                        # game_controller._log_debug(f"System info: {result}")
                # get processes
                if "Processes" in data:
                    # data[processes] = {'All'}|{['PID','PPID','User','Port']}
                    # the elements in Processes determines what data is required from the processes
                    procs_to_ignore = ['-']
                    # game_controller._log_debug(f'Getting Process info for {host}')
                    if 'All' in data['Processes'] or 'PID' in data['Processes']:
                        # try tasklist command for windows
                        result = game_controller.execute_ssh_command('tasklist /FO csv', host)

                        if result == '' or 'command not found' in result or 'Invalid command' in result:
                            # if failed then try linux command
                            result = game_controller.execute_ssh_command(
                                "ps -e -o user,pid,ppid,command | awk '{print $1,$2,$3,$4}'", host)
                            result = result.split('\n')

                            for proc in result:
                                if "Image Name" in proc or 'USER' in proc:
                                    continue
                                p = proc.split(' ')
                                if len(p) == 4:
                                    if p[0] != 'ec2-user' and p[0] != 'ubuntu':
                                        if '/' in p[3] and '[' not in p[3]:
                                            path, name = p[3][:p[3].rfind('/')], p[3][p[3].rfind('/') + 1:]
                                        else:
                                            name = p[3]
                                            path = None
                                        obs.add_process(hostid=host, pid=int(p[1]), parent_pid=int(p[2]),
                                                        process_name=name, path=path, username=p[0])
                                    else:
                                        procs_to_ignore.append(p[1])
                        else:
                            result = result.split('\n')[1:]
                            # game_controller._log_debug(f"Process info: {result}")
                            for r in result:
                                r = r.replace('"', '').split(',')
                                if len(r) > 1:
                                    obs.add_process(hostid=host, process_name=r[0], pid=r[1])
                            if 'All' in data['Processes'] or 'PPID' in data['Processes']:
                                result = game_controller.execute_ssh_command(
                                    "wmic process get processid,parentprocessid,name,executablepath /format:csv", host)
                                result = result.split('\n')[2:]
                                # game_controller._log_debug(f"Process info: {result}")
                                ssh_command = ''
                                if 'All' in data['Processes'] or 'User' in data['Processes']:
                                    for r in result:
                                        r = r.replace('"', '').split(',')
                                        if len(r) > 4:
                                            path = '\\'.join(r[1].split('\\')[:-1])
                                            ssh_command += f"""wmic process where "name='{r[2]}'" call GetOwner;"""

                                    result2 = game_controller.execute_ssh_command(ssh_command, host)
                                    count = 0
                                    # self._log_debug(f'results2: {result2[:2000]}')
                                    r2 = result2.split('{')[1:]
                                    for r in result:
                                        r = r.replace('"', '').split(',')
                                        if len(r) > 4:
                                            path = '\\'.join(r[1].split('\\')[:-1])
                                            if 'User = ' in r2[count]:
                                                # self._log_debug(f'r2: {r2[count]}')
                                                obs.add_process(hostid=host, path=path if path != '' else None,
                                                                process_name=r[2], pid=r[3], parent_pid=r[4],
                                                                username=r2[count].split('User = ')[1].split('"')[1])
                                            else:
                                                obs.add_process(hostid=host, path=path if path != '' else None,
                                                                process_name=r[2], pid=r[3], parent_pid=r[4])
                                            count += 1

                                else:
                                    for r in result:
                                        r = r.replace('"', '').split(',')
                                        path = '\\'.join(r[1].split('\\')[:-1])
                                        if len(r) > 4:
                                            obs.add_process(hostid=host, path=path if path != '' else None,
                                                            process_name=r[2], pid=r[3], parent_pid=r[4])

                    if 'All' in data['Processes'] or 'Port' in data['Processes']:
                        result = game_controller.execute_ssh_command(
                            "sudo netstat -npl4 | sed 's/LISTEN/ /g' | awk '{print $1,$4,$6}'", host)
                        if result == '' or 'command not found' in result or 'Invalid command' in result:
                            # linux command failed so try windows command
                            result2 = game_controller.execute_ssh_command(f"netstat -ano", host)
                            for line in result2.split('\n'):
                                l = re.sub(' +', ' ', line)
                                l = l.split(' ')
                                # self._log_debug(list(enumerate(l)))
                                if len(l) > 5 and l[4] == 'LISTENING':
                                    lp = int(l[2].split(':')[-1])
                                    la = ''.join(l[2].split(':')[:-1])
                                    ls = ''.join(l[3].split(':')[:-1])
                                    # self._log_debug(la)
                                    if ls == '0.0.0.0':
                                        if la == '127.0.0.1':
                                            obs.add_process(hostid=host, pid=int(l[5]), local_port=lp, local_address=la)
                                        else:
                                            obs.add_process(hostid=host, pid=int(l[5]), local_port=lp, local_address=ls)
                        else:
                            # linux command worked
                            for conn in result.split('\n')[2:]:
                                c = conn.split(' ')
                                if len(c) == 3 and '/' in c[2] and ':' in c[1]:
                                    pid = c[2].split('/')[0]
                                    if pid not in procs_to_ignore:
                                        obs.add_process(hostid=host, pid=int(pid), transport_protocol=c[0],
                                                        local_address=c[1].split(':')[0], local_port=c[1].split(':')[1])

                # get files
                if "Files" in data:
                    result = ''  # game_controller.execute_ssh_command('dir', host)
                    if result == '' or 'command not found' in result or 'Invalid command' in result:
                        # likely fail due to wrong OS
                        result = game_controller.execute_ssh_command('ls', host)
                    # game_controller._log_debug(f"File info: {result}")
                # get users
                if "User info" in data:
                    # game_controller._log_debug(f'Getting User info for {host}')

                    result = game_controller.execute_ssh_command('wmic useraccount get name,sid /format:csv', host)
                    # game_controller._log_debug(f'result: {result}')
                    if result == '' or 'command not found' in result or 'Invalid command' in result:
                        # likely fail due to wrong OS
                        result = game_controller.execute_ssh_command('cat /etc/passwd', host)
                        for r in result.split('\n'):
                            r = r.split(':')
                            if len(r) >= 4:
                                obs.add_user_info(hostid=host, username=r[0], uid=int(r[2]), gid=int(r[3]))
                        result = game_controller.execute_ssh_command('cat /etc/group', host)
                        for r in result.split('\n'):
                            r = r.split(':')
                            if len(r) == 4:
                                obs.add_user_info(hostid=host, group_name=r[0], gid=int(r[2]))
                                for u in r[3].split(','):
                                    if u != '':
                                        obs.add_user_info(hostid=host, group_name=r[0], gid=int(r[2]), username=u)
                    # game_controller._log_debug(f"User info: {result}")
                    else:
                        for r in result.split('\n')[2:]:
                            r = r.split(',')
                            if len(r) > 2:
                                obs.add_user_info(hostid=host, username=r[1], uid=r[2])
                        # game_controller._log_debug(f"User info: {result}")

                if "Interfaces" in data:
                    # game_controller._log_debug(f'Getting Interface info for {host}')

                    result = game_controller.execute_ssh_command('ip a', host)
                    # game_controller._log_debug(f'ip a on {host}: {result}')
                    if 'command not found' in result:
                        result = game_controller.execute_ssh_command('ipconfig', host)
                        # game_controller._log_debug(f"Interface: {result}")
                        interfaces = result.split('adapter')[1:]
                        for i in interfaces:
                            ip = None
                            mask = None
                            for line in i.split('\n'):
                                if 'IPv4 Address' in line and ': ' in line:
                                    ip = line.split(': ')[1]
                                if 'Subnet Mask' in line and ': ' in line:
                                    mask = line.split(': ')[1]
                            if ip is not None and mask is not None:
                                obs.add_interface_info(hostid=host,
                                                       ip_address=ip,
                                                       subnet=IPv4Network(f'{ip}/{mask}', strict=False))
                    else:
                        for line in result.split('\n'):
                            if 'inet ' in line:
                                l = re.sub(' +', ' ', line).split(' ')
                                if len(l) > 6:
                                    # game_controller._log_debug(f'hostid={host}, interface_name={l[8]}, ip_address={l[2].split("/")[0]}, subnet={IPv4Network(l[2], strict=False)}')
                                    obs.add_interface_info(hostid=host, interface_name=l[8],
                                                           ip_address=l[2].split("/")[0],
                                                           subnet=IPv4Network(l[2], strict=False))

                                else:
                                    # game_controller._log_debug(f'hostid={host}, interface_name={l[5]}, ip_address={l[2].split("/")[0]}, subnet={IPv4Network(l[2], strict=False)}')
                                    obs.add_interface_info(hostid=host, interface_name=l[5],
                                                           ip_address=l[2].split("/")[0],
                                                           subnet=IPv4Network(l[2], strict=False))

            return obs
        return game_controller.get_true_state(self.info)



# Copyright DST Group. Licensed under the MIT license.
from ipaddress import IPv4Address
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.open_connection import OpenConnection
from csle_cyborg.simulator.process import ProcessType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# smbclient -L //target -N
class SMBAnonymousConnection(OpenConnection):
    def __init__(self, ip_address: IPv4Address, session: int, agent: str):
        super().__init__(session=session, agent=agent)
        self.target = ip_address

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if not session.active or self.target not in state.ip_addresses:
            return obs

        # check if smbclient is on session's dict
        client = False
        for file in session.host.files:
            if file.name == "smbclient" and file.check_executable(session.user):
                client = True

        if not client:
            return obs

        # check if ip is on same subnet as dict of session
        shared_subnet = False
        for interface in session.host.interfaces:
            if interface.ip_address != IPv4Address("127.0.0.1") and self.target in interface.subnet.ip_addresses:
                shared_subnet = True
                break

        if not shared_subnet:
            return obs

        smb_proc = None
        for proc in state.hosts[state.ip_addresses[self.target]].processes:
            if proc.process_type == ProcessType.SMB:
                smb_proc = proc
                break

        if smb_proc is not None:
            for conn in smb_proc.connections:
                if conn['local_address'] == IPv4Address("0.0.0.0"):
                    obs.set_success(True)
                    obs.add_process(local_address=str(self.target), status="open", process_type="smb",
                                    process_version=smb_proc.version)
                    break
        else:
            obs.add_interface_info(ip_address=str(self.target))

        return obs
        # what else to check as far as obs is affected?
        # do we change the state at all? Might create an event that blue can see



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.remote_code_execution import RemoteCodeExecution


class RubyOnRails(RemoteCodeExecution):
    def __init__(self):
        super().__init__()

    def sim_execute(self, state):
        pass



from ipaddress import IPv4Address

from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.exploit_action import ExploitAction
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.process import Process
from csle_cyborg.simulator.state import State


class RemoteCodeExecutionOnSMTP(ExploitAction):
    def __init__(self, session: int, agent: str, ip_address: IPv4Address, target_session: int):
        super().__init__(session, agent, ip_address, target_session)

    def sim_execute(self, state: State) -> Observation:
        return self.sim_exploit(state, 443, 'webserver')

    def test_exploit_works(self, target_host: Host, vuln_proc: Process):
        # check if OS and process information is correct for exploit to work
        return "rfi" in vuln_proc.properties



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import MSFAction


class MeterpreterAction(MSFAction):
    def __init__(self, session: int, agent: str, target_session: int):
        super().__init__(session=session, agent=agent)
        self.meterpreter_session = target_session

    def __str__(self):
        return super(MeterpreterAction, self).__str__() + f", Meterpreter Session: {self.meterpreter_session}"



from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.state import State


class DensityScout(ConcreteAction):
    def __init__(self, session: int, agent: str, target_session: int):
        super(DensityScout, self).__init__(session=session, agent=agent)
        self.target_session = target_session

    def sim_execute(self, state: State) -> Observation:
        obs = Observation()
        if self.session not in state.sessions[self.agent] or self.target_session not in state.sessions[self.agent]:
            obs.set_success(False)
            return obs
        target_host: Host = state.hosts[state.sessions[self.agent][self.target_session].host]
        session = state.sessions[self.agent][self.session]
        target_session = state.sessions[self.agent][self.target_session]

        if not session.active or not target_session.active:
            obs.set_success(False)
            return obs
        obs.set_success(True)
        for file in target_host.files:
            obs.add_file_info(hostid=target_host.hostname, name=file.name, path=file.path, density=file.density)

        return obs


# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.internal_enumeration import InternalEnumeration


class NetworkInfoEnumeration(InternalEnumeration):
    def __init__(self):
        super().__init__()

    def sim_execute(self):
        pass



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.enums import QueryType
from csle_cyborg.shared.observation import Observation
from .velociraptor_action import VelociraptorAction


class GetOSInfo(VelociraptorAction):

    def __init__(self, session: int, agent: str):
        super().__init__(session=session,
                         query_type=QueryType.ASYNC,
                        agent=agent)
        self.agent = agent

    def sim_execute(self, state):
        obs = Observation()
        return obs



## The following code contains work of the United States Government and is not subject to domestic copyright protection under 17 USC § 105.
## Additionally, we waive copyright and related rights in the utilized code worldwide through the CC0 1.0 Universal public domain dedication.

"""
Implements misinformation actions for blue agents
"""
# pylint: disable=invalid-name
from random import choice
from typing import Tuple, List, Optional
from abc import ABC, abstractmethod
from dataclasses import dataclass
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.enums import DecoyType
from csle_cyborg.simulator.host import Host
from csle_cyborg.simulator.session import Session
from csle_cyborg.simulator.state import State


@dataclass
class Decoy:
    """
    Contains information necessary to create a misinform process on a host
    """
    service_name: str
    name: str
    open_ports: List[dict]
    process_type: str
    process_path: Optional[str] = None
    version: Optional[str] = None
    properties: Optional[List[str]] = None


def _is_host_using_port(host: Host, port: int):
    """
    Convenience method for checking if a host is using a port
    """
    if host.processes is not None:
        for proc in host.processes:
            for proc_state in proc.get_state():
                if proc_state.get('local_port', None) == port:
                    return True
    return False


class DecoyFactory(ABC):
    """
    Assembles process informationt to appear as a vulnerable process
    """

    @abstractmethod
    def make_decoy(self, host: Host) -> Decoy:
        """
        Creates a Decoy instance that contains the necessary information
        to put a decoy on a given host.

        :param host: Host that this decoy will be placed on
        """

    @abstractmethod
    def is_host_compatible(self, host: Host) -> bool:
        """
        Determines whether an instance of this decoy can be placed
        successfully on the given host

        :param host: Host to examine for compatibility with this decoy.
        """


class SSHDDecoyFactory(DecoyFactory):
    """
    Assembles process information to appear as an ssh server
    """

    def make_decoy(self, host: Host) -> Decoy:
        del host
        return Decoy(service_name="sshd", name="Sshd.exe",
                     open_ports=[{'local_port': 22, 'local_address': '0.0.0.0'}],
                     process_type="sshd",
                     process_path="C:\\Program Files\\OpenSSH\\usr\\sbin")

    def is_host_compatible(self, host: Host) -> bool:
        return not _is_host_using_port(host, 22)


sshd_decoy_factory = SSHDDecoyFactory()


class ApacheDecoyFactory(DecoyFactory):
    """
    Assembles process information to appear as an apache server
    """

    def make_decoy(self, host: Host) -> Decoy:
        del host
        return Decoy(service_name="apache2", name="apache2",
                     open_ports=[{'local_port': 80, 'local_address': '0.0.0.0'}],
                     process_type="webserver", properties=["rfi"],
                     process_path="/usr/sbin")

    def is_host_compatible(self, host: Host) -> bool:
        return not _is_host_using_port(host, 80)


apache_decoy_factory = ApacheDecoyFactory()


class SMSSDecoyFactory(DecoyFactory):
    """
    Assembles process information to appear as smss
    """

    def make_decoy(self, host: Host) -> Decoy:
        del host
        return Decoy(service_name="smss", name="Smss.exe",
                     open_ports=[{'local_port': 139, 'local_address': '0.0.0.0'}],
                     process_type="smss")

    def is_host_compatible(self, host: Host) -> bool:
        return not _is_host_using_port(host, 139)


smss_decoy_factory = SMSSDecoyFactory()


class TomcatDecoyFactory(DecoyFactory):
    """
    Assembles process information to appear as a tomcat server
    """

    def make_decoy(self, host: Host) -> Decoy:
        del host
        return Decoy(service_name="tomcat", name="Tomcat.exe",
                     open_ports=[{'local_port': 443, 'local_address': '0.0.0.0'}],
                     process_type="webserver", properties=["rfi"])

    def is_host_compatible(self, host: Host) -> bool:
        return not _is_host_using_port(host, 443)


tomcat_decoy_factory = TomcatDecoyFactory()


class SvchostDecoyFactory(DecoyFactory):
    """
    Assembles process information to appear as svchost
    """

    def make_decoy(self, host: Host) -> Decoy:
        del host
        return Decoy(service_name="svchost", name="Svchost.exe",
                     open_ports=[{'local_port': 3389, 'local_address': '0.0.0.0'}],
                     process_type="svchost")

    def is_host_compatible(self, host: Host) -> bool:
        return not _is_host_using_port(host, 3389)


svchost_decoy_factory = SvchostDecoyFactory()


class Misinform(Action):
    """
    Creates a misleading process on the designated host depending on
    available and compatible options.
    """

    def __init__(self, *, session: int, agent: str, hostname: str):
        self.agent = agent
        self.session = session
        self.hostname = hostname
        self.decoy_type = DecoyType.EXPLOIT
        self.candidate_decoys = (
            sshd_decoy_factory,
            apache_decoy_factory,
            smss_decoy_factory,
            tomcat_decoy_factory,
            svchost_decoy_factory)

    def emu_execute(self) -> Observation:
        raise NotImplementedError

    def sim_execute(self, state: State) -> Observation:
        obs_fail = Observation(False)
        obs_succeed = Observation(True)

        sessions = [s for s in state.sessions[self.agent].values() if
                    s.host == self.hostname]
        if len(sessions) == 0:
            return obs_fail

        session = choice(sessions)
        host = state.hosts[self.hostname]

        try:
            decoy_factory = self.__select_one_factory(host)
            decoy = decoy_factory.make_decoy(host)
            self.__create_process(obs_succeed, session, host, decoy)
            # print ("Misinform Success. Result: {}".format(result))

            return obs_succeed

        except RuntimeError:
            # print ("Misinform FAILURE")
            return obs_fail

    def __select_one_factory(self, host: Host) -> DecoyFactory:
        """
        Examines all decoy factories and returns one randomly compatible one.
        Raises RuntimeError if no compatible ones are found.
        """

        compatible_factories = [factory for factory in self.candidate_decoys
                                if factory.is_host_compatible(host)]

        if len(compatible_factories) == 0:
            raise RuntimeError("No compatible factory")

        return choice(list(compatible_factories))

    def __create_process(self, obs: Observation, sess: Session, host: Host,
                         decoy: Decoy) -> None:
        """
        Creates a process & service from Decoy on current host, adds it
        to the observation.
        """

        parent_pid = 1

        process_name = decoy.name
        username = sess.username
        version = decoy.version
        open_ports = decoy.open_ports
        process_type = decoy.process_type
        process_props = decoy.properties

        service_name = decoy.service_name

        new_proc = host.add_process(name=process_name, ppid=parent_pid,
                                    user=username, version=version, process_type=process_type,
                                    open_ports=open_ports, decoy_type=self.decoy_type,
                                    properties=process_props)

        host.add_service(service_name=service_name, process=new_proc.pid,
                         session=sess)

        obs.add_process(hostid=self.hostname, pid=new_proc.pid,
                        parent_pid=parent_pid, name=process_name,
                        username=username, service_name=service_name,
                        properties=process_props)

    def __str__(self):
        return f"{self.__class__.__name__} {self.hostname}"



# Copyright DST Group. Licensed under the MIT license.
import time
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.global_actions.global_action import GlobalAction


class GlobalSleep(GlobalAction):

    def __init__(self, t=1):
        super().__init__()
        self.t = t

    def emu_execute(self, team_server) -> Observation:
        obs = Observation()
        time.sleep(self.t)
        obs.set_success(True)
        obs.add_raw_obs("Sleeping...")
        return obs



import inspect
from csle_cyborg.agents.wrappers.base_wrapper import BaseWrapper


class ReduceActionSpaceWrapper(BaseWrapper):
    def __init__(self, env=None, agent=None):
        super().__init__(env, agent)
        self.action_signature = {}
        self.known_params = {}
        self.params_to_fix_at_start = ['port']
        self.fixed_size = {}

    def action_space_change(self, action_space: dict) -> dict:
        assert type(action_space) is dict, f"Wrapper required a dictionary action space. " \
                                           f"Please check that the wrappers below the ReduceActionSpaceWrapper return the action space as a dict"

        for param in self.params_to_fix_at_start:
            if param in self.fixed_size:
                remove_keys = []
                for p in action_space[param].keys():
                    if p not in self.fixed_size[param]:
                        remove_keys.append(p)

                for key in remove_keys:
                    action_space[param].pop(key)
                # action_space[param] = self.fixed_size[param]
            else:
                self.fixed_size[param] = list(action_space[param].keys())
        params = ['action']
        for action in action_space['action']:
            if action not in self.action_signature:
                self.action_signature[action] = inspect.signature(action).parameters
            for p in self.action_signature[action]:
                if p not in params:
                    params.append(p)
        to_remove = []
        for key, value in action_space.items():
            if key not in params:
                to_remove.append(key)

        for p in to_remove:
            action_space.pop(p)

        return action_space

    def get_attr(self,attribute:str):
        return self.env.get_attr(attribute)

    def reset(self, agent=None):
        self.fixed_size = {}
        return self.env.reset(agent)



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.shell_action import ShellAction
from csle_cyborg.shared.enums import FileType
from csle_cyborg.shared.observation import Observation


class FindFlag(ShellAction):

    def __init__(self, session, agent):
        super().__init__(session, agent)

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs

        session = state.sessions[self.agent][self.session]
        if not session.active:
            return obs

        for file in session.host.files:
            if file.file_type == FileType.FLAG:
                if file.check_readable(session.user):
                    obs.set_success(True)
                    obs.add_file_info(hostid="hostid0", path=file.path, name=file.name, file_type=file.file_type)
                    return obs

        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.shell_action import ShellAction
from csle_cyborg.shared.enums import OperatingSystemType
from csle_cyborg.shared.observation import Observation


class IPConfig(ShellAction):
    def __init__(self, session, agent):
        super().__init__(session, agent)

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs

        if state.sessions[self.agent][self.session].active:
            host = state.sessions[self.agent][self.session].host
            obs.add_system_info(hostid="hostid0", os_type=host.os_type)
            if host.os_type == OperatingSystemType.WINDOWS:
                obs.set_success(True)
                for interface in host.interfaces:
                    obs.add_interface_info(hostid="hostid0", **(interface.get_state()))
        return obs


# Copyright DST Group. Licensed under the MIT license.
import string
from ipaddress import IPv4Address
import random
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import MSFAction, lo
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.enums import AppProtocol
from csle_cyborg.shared.observation import Observation


class UpgradeToMeterpreter(MSFAction):
    def __init__(self, session: int, agent: str, target_session: int):
        super().__init__(session=session, agent=agent)
        self.session_to_upgrade = target_session

    def sim_execute(self, state):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent] or self.session_to_upgrade not in state.sessions[self.agent]:
            return obs
        server_session = state.sessions[self.agent][self.session]
        session_to_upgrade = state.sessions[self.agent][self.session_to_upgrade]

        # action fails if either chosen session is not active or of the correct type
        if server_session.session_type != SessionType.MSF_SERVER \
                or not (
                session_to_upgrade.session_type == SessionType.MSF_SHELL or session_to_upgrade.session_type == SessionType.METERPRETER) or not server_session.active \
                or not session_to_upgrade.active:
            return obs

        # find shared subnet of the two hosts
        server_interface = None
        up_interface = None
        # test if the two sessions are on the same host
        if server_session.host == session_to_upgrade.host:
            server_interface = server_session.host.get_interface(interface_name='lo')
        else:
            for upgrade_interface in state.hosts[session_to_upgrade.host].interfaces:
                if upgrade_interface.ip_address != lo:
                    server_session, server_interface = self.get_local_source_interface(local_session=server_session,
                                                                                       remote_address=upgrade_interface.ip_address,
                                                                                       state=state)
                if server_interface is not None:
                    up_interface = upgrade_interface
                    break

        if server_interface is None:
            return obs

        server_address = server_interface.ip_address
        upgrade_address = up_interface.ip_address

        obs.set_success(True)

        new_session = state.add_session(host=session_to_upgrade.host, agent=self.agent,
                                        user=session_to_upgrade.username, session_type="meterpreter",
                                        parent=server_session)
        process = state.hosts[new_session.host].get_process(new_session.pid)
        process.ppid = session_to_upgrade.pid
        process.path = "/tmp/"
        # Randomly generate name:
        process.name = ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase) for _ in range(5))
        local_port = state.hosts[session_to_upgrade.host].get_ephemeral_port()
        new_connection = {"Application Protocol": AppProtocol.TCP,
                          "remote_address": server_address,
                          "remote_port": 4433,
                          "local_address": upgrade_address,
                          "local_port": local_port}
        process.connections.append(new_connection)

        remote_port = {"local_port": 4433,
                       "Application Protocol": AppProtocol.TCP,
                       "local_address": server_address,
                       "remote_address": upgrade_address,
                       "remote_port": local_port
                       }
        state.hosts[server_session.host].get_process(server_session.pid).connections.append(remote_port)

        obs.add_session_info(hostid=str(self.session_to_upgrade), session_id=new_session.ident,
                             session_type=new_session.session_type, agent=self.agent)

        obs.add_process(hostid=str(server_address), local_address=server_address, local_port=4433,
                        remote_address=upgrade_address,
                        remote_port=local_port)
        obs.add_process(hostid=str(self.session_to_upgrade), local_address=upgrade_address, local_port=local_port,
                        remote_address=server_address,
                        remote_port=4433)
        return obs

    def emu_execute(self, session_handler) -> Observation:
        obs = Observation()
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs
        output = session_handler.execute_module(mtype='post', mname='multi/manage/shell_to_meterpreter',
                                                opts={'SESSION': self.session_to_upgrade})
        obs.add_raw_obs(output)
        obs.set_success(False)
        for line in output.split('\n'):
            if '[*] Meterpreter session' in line:
                obs.set_success(True)
                split = line.split(' ')
                # print(list(enumerate(split)))
                session = int(split[3])
                remote_address, remote_port = split[5][1:].split(':')
                local_address, local_port = split[7][:-1].split(':')
                # date = datetime.fromisoformat(split[10] + ' ' + split[11])
                obs.add_session_info(hostid=str(self.session_to_upgrade), session_id=session, agent=self.agent,
                                     session_type='meterpreter')
                obs.add_process(hostid=str(self.session_to_upgrade), local_port=local_port, remote_port=remote_port,
                                local_address=local_address, remote_address=remote_address)
                obs.add_process(hostid=str(remote_address), remote_port=local_port, local_port=remote_port,
                                remote_address=local_address, local_address=remote_address)
                # print(f'session: {session}')
                # print(f'local_port: {local_port}')
                # print(f'local_address: {local_address}')
                # print(f'remote_port: {remote_port}')
                # print(f'remote_address: {remote_address}')
                # print(f'date: {date}')
        '''Example obs
        [*] Upgrading session ID: 1
        [*] Starting exploit/multi/handler
        [*] Started reverse TCP handler on 10.0.20.245:4433 
        [*] Sending stage (985320 bytes) to 10.0.2.164
        [*] Meterpreter session 2 opened (10.0.20.245:4433 -> 10.0.2.164:38182) at 2020-08-03 06:10:00 +0000
        [*] Command stager progress: 100.00% (773/773 bytes)
        [*] Post module execution completed'''
        return obs

    def __str__(self):
        return super(UpgradeToMeterpreter, self).__str__() + f", Shell Session: {self.session_to_upgrade}"



# Copyright DST Group. Licensed under the MIT license.

from csle_cyborg.shared.enums import QueryType
from csle_cyborg.shared.observation import Observation
from .velociraptor_action import VelociraptorAction


class GetProcessListForUsername(VelociraptorAction):
    """Get a list of the process IDs that are running for a user"""

    def __init__(self, session: int, agent: str, hostname: str, username: str):
        super().__init__(session=session,
                         query_type=QueryType.ASYNC,
                         agent=agent)
        self.agent = agent
        self.hostname = hostname
        self.username = username
        self.parameters['artifactname'] = 'Custom.CybORG.Generic.System.Pslist'
        self.query = (
            "select collect_client("
            "client_id='<host_id>', "
            "artifacts=['Custom.CybORG.Generic.System.Pslist'], "
            "env=dict(agentid='<agentid>', "
            "artifactname='<artifactname>', "
            "tag='<tag>')) FROM scope()"
        )
        self.source_query = (
            "select Pid, Username from source("
            "flow_id='<flowid>', "
            "client_id='<host_id>', "
            "artifact='<artifactname>') "
            "where '<username>' in Username"
        )

    def sim_execute(self, state):
        raise NotImplementedError

    def parse(self, results: list) -> Observation:
        """Parses the results of the execute action to create an observation"""
        obs = Observation()

        if results == []:
            obs.set_success(False)
            return obs
        else:
            obs.set_success(True)

        obs.add_system_info(hostid='0', hostname=self.hostname)

        for p in results:
            if p['Username'] == self.username:
                obs.add_process(
                    hostid='0',
                    pid=p['Pid'],
                    username=self.username
                )
            else:
                user = p['Username'].rsplit('\\')
                if len(user) == 2 and user[1] == self.username:
                    obs.add_process(
                        hostid='0',
                        pid=p['Pid'],
                        username=self.username
                    )
        return obs



from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
import random
from ipaddress import IPv4Address
from csle_cyborg.shared.actions.velociraptor_actions_folder.velociraptor_poll import VelociraptorPoll
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.credential_access_folder.ssh_access import SSHAccess
from csle_cyborg.shared.actions.shell_actions_folder.kill_process_linux import KillProcessLinux


class CounterKillchainAgent(BaseAgent):

    def __init__(self, action_size=None, state_size=None):
        self.colour = 'Blue'
        self.killchains = {
            'kill_process': [VelociraptorPoll, SSHAccess, VelociraptorPoll, KillProcessLinux,
                             VelociraptorPoll, KillProcessLinux]
        }

        self.action_parameters = {
            'Sleep': {},
            #             'IFConfig':{'agent':'Blue','session':0},
            #             'IPConfig':{'agent':'Blue','session':0},
            #             'FindFlag':{'agent':'Blue','session':0},
            'SSHAccess': {'agent': 'Blue', 'session': 1, 'username': 'blue', 'password': 'blue_knight',
                          'ip_address': IPv4Address('10.0.2.1'), 'port': 22},
            #             'PingSweep':{'agent':'Blue','session':0},
            #             'NmapScan':{'agent':'Blue','session':0},
            #             'DeleteFileLinux':{'agent':'Blue','session':0},
            #             'DeleteFileWindows':{'agent':'Blue','session':0},
            'KillProcessLinux': {'agent': 'Blue', 'session': 4, 'process': 0},
            #             'KillProcessWindows':{'agent':'Blue','session':0},
            #             'AddUserWindows':{'agent':'Blue','session':0},
            #             'AddUserLinux':{'agent':'Blue','session':0},
            #             'DisableUserWindows':{'agent':'Blue','session':0},
            #             'DisableUserLinux':{'agent':'Blue','session':0},
            #             'RemoveUserFromGroupWindows':{'agent':'Blue','session':0},
            #             'RemoveUserFromGroupLinux':{'agent':'Blue','session':0},
            #             'SMBAnonymousConnection':{'agent':'Blue','session':0},
            #             'ShellPS':{'agent':'Blue','session':0},
            #             'SystemInfo':{'agent':'Blue','session':0},
            #             'Uname':{'agent':'Blue','session':0},
            #             'NetcatConnect':{'agent':'Blue','session':0},
            #             'SSHHydraBruteForce':{'agent':'Blue','session':0},
            #             'ReadShadowFile':{'agent':'Blue','session':0},
            #             'ReadPasswdFile':{'agent':'Blue','session':0},
            #             'StopService':{'agent':'Blue','session':0},
            #             'StartService':{'agent':'Blue','session':0},
            #             'GetFileInfo':{'agent':'Blue','session':0},
            #             'GetLocalGroups':{'agent':'Blue','session':0},
            #             'GetOSInfo':{'agent':'Blue','session':0},
            #             'GetProcessInfo':{'agent':'Blue','session':0},
            #             'GetProcessList':{'agent':'Blue','session':0},
            #             'GetUsers':{'agent':'Blue','session':0},
            'VelociraptorPoll': {'agent': 'Blue', 'session': 0}
        }

        self.kchoice = random.choice(list(self.killchains.keys()))
        self.killchain = self.killchains[self.kchoice]
        self.count = 0
        self.memory = []

    def train(self, results):
        pass

    def set_initial_values(self, action_space, observation):
        pass

    def get_action(self, observation, action_space):
        pid = self._process_observation(observation)
        if pid is not None:
            self.action_parameters['KillProcessLinux']['process'] = pid
        position = self.count % len(self.killchain)
        action_class = self.killchain[position]
        action_params = self.action_parameters[action_class.__name__]
        self.count += 1
        return action_class(**action_params)

    def _process_observation(self, observation):
        if self.count == 1:
            self.memory = observation['PublicFacing']['Processes']
        else:
            try:
                obs = observation['PublicFacing']['Processes']
            except:
                return None
            for process in obs:
                if (process not in self.memory) \
                        and ('Connections' in process.keys()) \
                        and (process['Username'] != 'blue'):
                    return process['PID']

    def end_episode(self):
        pass



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_action import MeterpreterAction
from csle_cyborg.shared.enums import OperatingSystemType, SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# Call ps from a meterpreter session - gives a list of processes with PID, name, user and path
class MeterpreterPS(MeterpreterAction):
    def __init__(self, session: int, agent: str):
        super().__init__(session=session, agent=agent)

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.METERPRETER or not session.active:
            return obs

        obs.set_success(True)
        users = []
        proc_sh = session.host.add_process(name="sh", user=session.user, path="/bin/")
        proc_ps = session.host.add_process(name="ps", user=session.user)

        if session.host.os_type == OperatingSystemType.LINUX:
            root = False
            for group in session.user.groups:
                if group.uid == 0:
                    root = True
                # Is this the best way to check this? Can group name be changed, is there some other way to check if it
                # is a user with minimal privileges?
                elif group.name == "nogroup":
                    for proc in session.host.processes:
                        if proc.user.username not in users:
                            users.append(proc.user.username)
                        obs.add_process(hostid="0", pid=proc.pid, process_name=proc.name,
                                        username=proc.user.username, path=proc.path)
                    for user in users:
                        obs.add_user_info(hostid="0", username=user)
                    # Need to be able to remove processes as well - remove /bin/sh and ps processes created above?
                    state.remove_process(host=session.host.hostname, pid=proc_sh.pid)
                    state.remove_process(host=session.host.hostname, pid=proc_ps.pid)
                    return obs

            obs.add_system_info(hostid="0", architecture=session.host.architecture)
            if root:
                state.remove_process(host=session.host.hostname, pid=proc_sh.pid)
                state.remove_process(host=session.host.hostname, pid=proc_ps.pid)
                for proc in session.host.processes:
                    if proc.user.username not in users:
                        users.append(proc.user.username)
                    obs.add_process(hostid="0", pid=proc.pid, process_name=proc.name,
                                    username=proc.user.username, parent_pid=proc.ppid, path=proc.path)
            else:
                state.remove_process(host=session.host.hostname, pid=proc_sh.pid)
                state.remove_process(host=session.host.hostname, pid=proc_ps.pid)
                for proc in session.host.processes:
                    if proc.user is not None and proc.user.username not in users:
                        users.append(proc.user.username)
                    obs.add_process(hostid="0", pid=proc.pid, process_name=proc.name,
                                    username=proc.user.username, parent_pid=proc.ppid)
            for user in users:
                obs.add_user_info(hostid="0", username=user)
            return obs

        else:
            state.remove_process(host=session.host.hostname, pid=proc_sh.pid)
            state.remove_process(host=session.host.hostname, pid=proc_ps.pid)
            obs.add_system_info(hostid="0", architecture=session.host.architecture)
            for proc in session.host.processes:
                if proc.user is not None and proc.user.username not in users:
                    users.append(proc.user.username)
                obs.add_process(hostid="0", pid=proc.pid, process_name=proc.name,
                                username=proc.user.username, parent_pid=proc.ppid, path=proc.path)
            for user in users:
                obs.add_user_info(hostid="0", username=user)
            return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.actions.shell_actions_folder.shell_action import ShellAction
from csle_cyborg.shared.enums import OperatingSystemType, SessionType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


# Call 'ps -o user,uid,pid,ppid,command ax' from a shell or msf shell session
# gives a list of processes with user, UID, PID, PPID, name, and path
# Note other potentially useful params after -o include: group, gid, eip, esp, stat
class ShellPS(ShellAction):
    def __init__(self, session: int, agent: str):
        super().__init__(session=session, agent=agent)

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.session in state.sessions[self.agent]:
            session = state.sessions[self.agent][self.session]

            if session.active:
                if session.host.os_type != OperatingSystemType.LINUX:
                    return obs

                if session.session_type != SessionType.SHELL and session.session_type != SessionType.MSF_SHELL:
                    return obs

                obs.add_system_info(hostid="0", os_type="linux")

                obs.set_success(True)
                users = []
                proc_ps = session.host.add_process(name="ps", user=session.user)

                for proc in session.host.processes:
                    if proc.user.username not in users:
                        users.append(proc.user)
                    obs.add_process(hostid="0", pid=proc.pid, process_name=proc.name,
                                    username=proc.user.username, parent_pid=proc.ppid, path=proc.path)

                for user in users:
                    obs.add_user_info(hostid="0", username=user.username, uid=user.uid)
                state.remove_process(host=session.host.hostname, pid=proc_ps.pid)

        return obs



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.shared.actions.agent_actions.agent_action import AgentAction


class CreateSession(AgentAction):
    """Create a new session on Team server for given game and agent.

    This session is designed for terminal commands executed on team server
    machine that affect a single game.

    Parameters
    ----------
    session_handler_cls : SessionHandler
        the session handler class to use for session
    handler_kwargs : dict, optional
        keyword arguments that will be passed to the session (default=None)
    """

    def __init__(self, session_handler_cls, handler_kwargs: dict = None):
        super().__init__()
        self.session_handler_cls = session_handler_cls
        self.handler_kwargs = {} if handler_kwargs is None else handler_kwargs

    def emu_execute(self, agent, *args, **kwargs) -> Observation:
        self._log_debug(f"Executing action: {self}")
        session_handler = self.session_handler_cls.create_new_session(
            **self.handler_kwargs
        )
        agent.add_session(session_handler)
        obs = Observation()
        obs.set_success(True)
        obs.add_session_info(agent=agent.agent_name,
                             **session_handler.get_info())
        return obs

    def __str__(self):
        return (f"{self.__class__.__name__}: "
                f"SessionHandler: {self.session_handler_cls.__name__}"
                f"kwargs: {self.handler_kwargs}")



# Copyright DST Group. Licensed under the MIT license.
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


class Action:

    def sim_execute(self, state: State) -> Observation:
        raise NotImplementedError

    def emu_execute(self) -> Observation:
        raise NotImplementedError

    def __str__(self):
        return f"{self.__class__.__name__}"

    def get_params(self) -> dict:
        return {key:value for key, value in self.__dict__.items() if not key.startswith('__') and not callable(key)}

    @property
    def cost(self):
        return 0


class Sleep(Action):
    def sim_execute(self, state):
        return Observation()

    def emu_execute(self) -> Observation:
        return Observation()


class InvalidAction(Action):
    def __init__(self, action: Action):
        self.action = action

    def sim_execute(self, state):
        return Observation(success=False)

    def emu_execute(self) -> Observation:
        return Observation(success=False)

    @property
    def cost(self):
        return -0.1



import random
from csle_cyborg.agents.simple_agents.base_agent import BaseAgent
from csle_cyborg.shared.results import Results
from csle_cyborg.shared.actions.abstract_actions.privilege_escalate import PrivilegeEscalate
from csle_cyborg.shared.actions.abstract_actions.exploit_remote_service import ExploitRemoteService
from csle_cyborg.shared.actions.abstract_actions.discover_remote_systems import DiscoverRemoteSystems
from csle_cyborg.shared.actions.abstract_actions.impact import Impact
from csle_cyborg.shared.actions.abstract_actions.discover_network_services import DiscoverNetworkServices


class B_lineAgent(BaseAgent):
    def __init__(self):
        self.action = 0
        self.target_ip_address = None
        self.last_subnet = None
        self.last_ip_address = None
        self.action_history = {}
        self.jumps = [0, 1, 2, 2, 2, 2, 5, 5, 5, 5, 9, 9, 9, 12, 13]

    def train(self, results: Results):
        """allows an agent to learn a policy"""
        pass

    def get_action(self, observation, action_space):
        # print(self.action)
        """gets an action from the agent that should be performed based on the agent's internal state and provided observation and action space"""
        session = 0

        while True:
            if observation['success'] == True:
                self.action += 1 if self.action < 14 else 0
            else:
                self.action = self.jumps[self.action]

            if self.action in self.action_history:
                action = self.action_history[self.action]

            # Discover Remote Systems
            elif self.action == 0:
                self.initial_ip = observation['User0']['Interface'][0]['IP Address']
                self.last_subnet = observation['User0']['Interface'][0]['Subnet']
                action = DiscoverRemoteSystems(session=session, agent='Red', subnet=self.last_subnet)
            # Discover Network Services- new IP address found
            elif self.action == 1:
                hosts = [value for key, value in observation.items() if key != 'success']
                get_ip = lambda x: x['Interface'][0]['IP Address']
                interfaces = [get_ip(x) for x in hosts if get_ip(x) != self.initial_ip]
                self.last_ip_address = random.choice(interfaces)
                action = DiscoverNetworkServices(session=session, agent='Red', ip_address=self.last_ip_address)

            # Exploit User1
            elif self.action == 2:
                action = ExploitRemoteService(session=session, agent='Red', ip_address=self.last_ip_address)

            # Privilege escalation on User Host
            elif self.action == 3:
                hostname = \
                [value for key, value in observation.items() if key != 'success' and 'System info' in value][0][
                    'System info']['Hostname']
                action = PrivilegeEscalate(agent='Red', hostname=hostname, session=session)

            # Discover Network Services- new IP address found
            elif self.action == 4:
                self.enterprise_host = [x for x in observation if 'Enterprise' in x][0]
                self.last_ip_address = observation[self.enterprise_host]['Interface'][0]['IP Address']
                action = DiscoverNetworkServices(session=session, agent='Red', ip_address=self.last_ip_address)

            # Exploit- Enterprise Host
            elif self.action == 5:
                self.target_ip_address = \
                [value for key, value in observation.items() if key != 'success'][0]['Interface'][0]['IP Address']
                action = ExploitRemoteService(session=session, agent='Red', ip_address=self.target_ip_address)

            # Privilege escalation on Enterprise Host
            elif self.action == 6:
                hostname = \
                [value for key, value in observation.items() if key != 'success' and 'System info' in value][0][
                    'System info']['Hostname']
                action = PrivilegeEscalate(agent='Red', hostname=hostname, session=session)

            # Scanning the new subnet found.
            elif self.action == 7:
                self.last_subnet = observation[self.enterprise_host]['Interface'][0]['Subnet']
                action = DiscoverRemoteSystems(subnet=self.last_subnet, agent='Red', session=session)

            # Discover Network Services- Enterprise2
            elif self.action == 8:
                self.target_ip_address = \
                [value for key, value in observation.items() if key != 'success'][2]['Interface'][0]['IP Address']
                action = DiscoverNetworkServices(session=session, agent='Red', ip_address=self.target_ip_address)

            # Exploit- Enterprise2
            elif self.action == 9:
                self.target_ip_address = \
                [value for key, value in observation.items() if key != 'success'][0]['Interface'][0]['IP Address']
                action = ExploitRemoteService(session=session, agent='Red', ip_address=self.target_ip_address)

            # Privilege escalation on Enterprise2
            elif self.action == 10:
                hostname = \
                [value for key, value in observation.items() if key != 'success' and 'System info' in value][0][
                    'System info']['Hostname']
                action = PrivilegeEscalate(agent='Red', hostname=hostname, session=session)

            # Discover Network Services- Op_Server0
            elif self.action == 11:
                action = DiscoverNetworkServices(session=session, agent='Red',
                                                 ip_address=observation['Op_Server0']['Interface'][0]['IP Address'])

            # Exploit- Op_Server0
            elif self.action == 12:
                info = [value for key, value in observation.items() if key != 'success']
                if len(info) > 0:
                    action = ExploitRemoteService(agent='Red', session=session,
                                                  ip_address=info[0]['Interface'][0]['IP Address'])
                else:
                    self.action = 0
                    continue
            # Privilege escalation on Op_Server0
            elif self.action == 13:
                action = PrivilegeEscalate(agent='Red', hostname='Op_Server0', session=session)
            # Impact on Op_server0
            elif self.action == 14:
                action = Impact(agent='Red', session=session, hostname='Op_Server0')

            if self.action not in self.action_history:
                self.action_history[self.action] = action
            return action

    def end_episode(self):
        self.action = 0
        self.target_ip_address = None
        self.last_subnet = None
        self.last_ip_address = None
        self.action_history = {}

    def set_initial_values(self, action_space, observation):
        pass



# Copyright DST Group. Licensed under the MIT license.
from datetime import datetime
from ipaddress import IPv4Address


# use msf module auxiliary/scanner/ssh/ssh_login
# set RHOSTS to target, maybe set VERBOSE to false, also can set RPORT but this is 22 by default
# set USERPASS_FILE (I set it to "/usr/share/metasploit_framework/data/wordlists/root_userpass.txt", having first added
# credentials "user user" to last line of file)
# Gives non-TTY session that needs to be interacted with first with "sessions -i 1" (may not be 1)
from time import strptime

from csle_cyborg.shared.actions.msf_actions_folder.msf_action import lo, lo_subnet
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.remote_code_execution import RemoteCodeExecution
from csle_cyborg.simulator.session import SessionType
from csle_cyborg.shared.enums import ProcessType
from csle_cyborg.shared.observation import Observation
from csle_cyborg.simulator.state import State


class SSHLoginExploit(RemoteCodeExecution):
    def __init__(self, ip_address: IPv4Address, agent: str, session: int, port: int):
        super().__init__(session=session, agent=agent)
        self.target = ip_address
        self.port = port

    def sim_execute(self, state: State):
        obs = Observation()
        obs.set_success(False)
        if self.session not in state.sessions[self.agent]:
            return obs
        session = state.sessions[self.agent][self.session]

        if session.session_type != SessionType.MSF_SERVER or not session.active:
            return obs

        target_subnet = None
        if self.target == lo:
            target_subnet = lo_subnet
            server_interface = [i for i in state.hosts[session.host].interfaces if i.ip_address == lo][0]
        else:
            for subnet in state.subnets.values():
                if self.target in subnet.ip_addresses:
                    target_subnet = subnet
                    break

            server_session, server_interface = self.get_local_source_interface(local_session=session, remote_address=self.target,
                                                                      state=state)


            if server_interface is None:
                return obs

            if not self.test_nacl(port=self.port, target_subnet=target_subnet,
                                  originating_subnet=state.subnets[server_interface.subnet]):
                return obs

        server_address = server_interface.ip_address
        if server_address is None:
            return obs

        if self.target == IPv4Address("127.0.0.1"):
            target_host = state.hosts[session.host]
        else:
            target_host = state.hosts[state.ip_addresses[self.target]]

        ssh_proc = None
        port = None
        # should also check for a non-ssh process listening on port 22 - only extra info in obs will be that this
        # process exists on that port and IP
        for proc in target_host.processes:
            if proc.process_type == ProcessType.SSH:
                for conn in proc.connections:
                    if conn['local_port'] == self.port:
                        ssh_proc = proc
                        port = conn
                if ssh_proc is not None:
                    break

        if ssh_proc is None or (port["local_address"] != IPv4Address("0.0.0.0") and port["local_address"] != self.target):
            return obs

        user_found = False
        user_user = None
        for u in target_host.users:
            if u.bruteforceable:
                user_found = True
                user_user = u

        if user_found:
            obs.set_success(True)
            obs.add_process(hostid=str(self.target), local_address=self.target, local_port=self.port, status="open",
                            process_type="ssh", app_protocol='ssh')

            user_ssh = target_host.add_process(name="sshd", ppid=ssh_proc.pid, path=ssh_proc.path,
                                                                   user=user_user, process_type="ssh")

            new_session = state.add_session(host=target_host.hostname, agent=self.agent,
                                            user=user_user.username, session_type="msf shell", parent=server_session.ident)
            process = target_host.get_process(new_session.pid)
            process.name = "bash"
            process.ppid = user_ssh.pid
            process.path = "/bin/"
            process.user = user_user
            remote_port = state.hosts[server_session.host].get_ephemeral_port()
            new_connection = {"local_port": self.port,
                              "Application Protocol": "tcp",
                              "remote_address": server_address,
                              "remote_port": remote_port,
                              "local_address": self.target}
            process.connections.append(new_connection)

            remote_port_dict = {'local_port': remote_port,
                                "Application Protocol": "ssh",
                                "local_address": server_address,
                                "remote_address": self.target,
                                "remote_port": self.port
                                }
            state.hosts[server_session.host].get_process(server_session.pid).connections.append(remote_port_dict)
            if session != server_session:
                remote_port = None
            obs.add_process(hostid=str(server_address), local_address=server_address, remote_address=str(self.target),
                            local_port=remote_port, remote_port=self.port)
            obs.add_process(hostid=str(self.target), local_address=str(self.target), remote_address=server_address,
                            local_port=self.port, remote_port=remote_port)
            obs.add_session_info(hostid=str(self.target), username=user_user.username, session_id=new_session.ident, session_type="msf shell", agent=self.agent)
            if target_host.os_type == OperatingSystemType.LINUX:
                obs.add_user_info(hostid=str(self.target), username=user_user.username, password=user_user.password, uid=user_user.uid)

                obs.add_system_info(hostid=str(self.target), hostname=target_host.hostname, architecture=target_host.architecture, os_kernel=target_host.kernel, os_type=target_host.os_type, os_distribution=target_host.distribution)
            else:
                obs.add_user_info(hostid=str(self.target), username=user_user.username, password=user_user.password)
        return obs

    def emu_execute(self, session_handler) -> Observation:
        obs = Observation()
        stop_on_success = True  # for the speed up during testing
        from csle_cyborg.CybORG import MSFSessionHandler
        if type(session_handler) is not MSFSessionHandler:
            obs.set_success(False)
            return obs
        output = session_handler.execute_module(mtype='auxiliary', mname='scanner/ssh/ssh_login',
                                         opts={'RHOSTS': str(self.target), "USERPASS_FILE": '/usr/share/wordlists/top100_userpass_msf.txt', "STOP_ON_SUCCESS": stop_on_success})
        obs.add_raw_obs(output)
        obs.set_success(False)
        username = None  # cheat to allow us to know the username of the new session
        try:
            for line in output.split('\n'):
                # Example success line: "[+] 10.0.2.164:22 - Success: 'pi:raspberry' 'uid=1001(pi) gid=1001(pi) groups=1001(pi) Linux pretend-pi 4.15.0-1057-aws #59-Ubuntu SMP Wed Dec 4 10:02:00 UTC 2019 x86_64 x86_64 x86_64 GNU/Linux '"
                # another example: [+] Success: 'pi:raspberry' 'uid=1001(pi) gid=1001(pi) groups=1001(pi) Linux ip-10-0-31-50 4.15.0-1057-aws #59-Ubuntu SMP Wed Dec 4 10:02:00 UTC 2019 x86_64 x86_64 x86_64 GNU/Linux '
                # another example: [+] Success: 'pi:raspberry' 'Could not chdir to home directory /home/pi: No such file or directory uid=1001(pi) gid=1001(pi) groups=1001(pi) Could not chdir to home directory /home/pi: No such file or directory Linux ip-10-0-10-199 4.15.0-1057-aws #59-Ubuntu SMP Wed Dec 4 10:02:00 UTC 2019 x86_64 x86_64 x86_64 GNU/Linux '
                if '[+]' in line:
                    # session_handler._log_debug(f"New positive result: {line}")
                    if ':' in line.split('Success: ')[0]:
                        ip_address, port = line.split('Success: ')[0].split(' ')[1].split(':')
                    else:
                        ip_address = self.target
                        port = self.port
                    line = line.split('Success: ')[-1]
                    split = line.split(' ')
                    # print(list(enumerate(split)))
                    if 'Could not chdir to home directory' in line:
                        split = [split[0]] + split[13:16] + split[28:]
                    username, password = split[0].replace('\'', '').split(':')
                    if 'id: command not found' not in line:
                        uid = split[1].replace('\'', '').split('=')[1].split('(')[0]
                        # gid = split[2].replace('\'', '').split('=')[1].split('(')[0]
                        # group = split[3].replace('\'', '').split('=')[1].split('(')[1].split(')')[0]
                        os = split[4]
                        hostname = split[5]
                        os_kv = split[6]
                        os_d = split[7].split('-')[1]
                        arch = split[15]

                        obs.add_user_info(hostid=str(self.target), username=username, uid=int(uid), password=password)
                        obs.add_process(hostid=str(self.target), local_port=port, local_address=ip_address, app_protocol='ssh', status='open', process_type="ssh")
                        obs.add_system_info(hostid=str(self.target), hostname=hostname, os_type=os, os_distribution=os_d, architecture=arch)
                    else:
                        obs.add_user_info(hostid=str(self.target), username=username, password=password)
                        obs.add_process(hostid=str(self.target), local_port=port, local_address=ip_address,
                                        app_protocol='ssh', status='open', process_type="ssh")
                    # print(f'ip: {ip_address}')
                    # print(f'port: {port}')
                    # print(f'user: {username}')
                    # print(f'password: {password}')
                    # print(f'uid: {uid}')
                    # print(f'gid: {gid}')
                    # print(f'groups: {group}')
                    # print(f'OS Type: {os}')
                    # print(f'hostname: {hostname}')
                    # print(f'OS kernel version: {os_kv}')
                    # print(f'OS distro: {os_d}')
                    # print(f'arch: {arch}')
                    # print(f'time: {time}')
                if '[*]' in line:
                    if "Command shell session" in line:
                        obs.set_success(True)
                        split = line.split(' ')
                        # print(list(enumerate(split)))
                        session = int(split[4])
                        local_address, local_port = split[8][:-1].split(':')
                        if '-' in split[6]:
                            temp = split[6].replace('(', '').split(':')[0]
                            origin, remote_address = temp.split('-')
                            # obs.add_process(hostid=str(origin), remote_address=remote_address, local_address=origin)
                            # obs.add_process(hostid=str(remote_address), local_address=remote_address, remote_address=origin)
                            remote_port = None
                        else:
                            remote_address, remote_port = split[6].replace('(', '').split(':')

                        #date = datetime.fromisoformat(split[10] + ' ' + split[11])
                        obs.add_session_info(hostid=str(self.target), username=username, session_id=session, agent=self.agent, session_type='msf shell')
                        obs.add_process(hostid=str(self.target), local_port=local_port, remote_port=remote_port, local_address=local_address, remote_address=remote_address)
                        obs.add_process(hostid=str(remote_address), local_address=remote_address,
                                        local_port=remote_port, remote_port=local_port,
                                        remote_address=local_address)
                        # print(f'session: {session}')
                        # print(f'local_port: {local_port}')
                        # print(f'local_address: {local_address}')
                        # print(f'remote_port: {remote_port}')
                        # print(f'remote_address: {remote_address}')
                        # print(f'date: {date}')
        except Exception as ex:
            session_handler._log_debug(f'Error occured in parsing of output: {output}')
            raise ex

        # session_handler._log_debug(output)
        return obs

    def __str__(self):
        return super(SSHLoginExploit, self).__str__() + f", Target: {self.target}:{self.port}"


from csle_cyborg.shared.actions.action import Sleep, InvalidAction
from csle_cyborg.shared.actions.abstract_actions.remove import Remove
from csle_cyborg.shared.actions.abstract_actions.analyse import Analyse
from csle_cyborg.shared.actions.abstract_actions.discover_network_services import DiscoverNetworkServices
from csle_cyborg.shared.actions.abstract_actions.discover_remote_systems import DiscoverRemoteSystems
from csle_cyborg.shared.actions.abstract_actions.exploit_remote_service import ExploitRemoteService
from csle_cyborg.shared.actions.abstract_actions.impact import Impact
from csle_cyborg.shared.actions.abstract_actions.misinform import Misinform
from csle_cyborg.shared.actions.abstract_actions.monitor import Monitor
from csle_cyborg.shared.actions.abstract_actions.privilege_escalate import PrivilegeEscalate
from csle_cyborg.shared.actions.abstract_actions.remove import Remove
from csle_cyborg.shared.actions.abstract_actions.restore import Restore
from csle_cyborg.shared.actions.agent_actions.agent_action import AgentAction
from csle_cyborg.shared.actions.agent_actions.agent_sleep import AgentSleep
from csle_cyborg.shared.actions.agent_actions.create_session import CreateSession
from csle_cyborg.shared.actions.agent_actions.get_initial_agent_observation import GetInitialAgentObservation
from csle_cyborg.shared.actions.agent_actions.list_sessions import ListSessions
from csle_cyborg.shared.actions.concrete_actions.blue_keep import BlueKeep
from csle_cyborg.shared.actions.concrete_actions.concrete_action import ConcreteAction
from csle_cyborg.shared.actions.concrete_actions.decoy_apache import DecoyApache
from csle_cyborg.shared.actions.concrete_actions.decoy_femitter import DecoyFemitter
from csle_cyborg.shared.actions.concrete_actions.decoy_haraka_smpt import DecoyHarakaSMPT
from csle_cyborg.shared.actions.concrete_actions.decoy_smss import DecoySmss
from csle_cyborg.shared.actions.concrete_actions.decoy_sshd import DecoySSHD
from csle_cyborg.shared.actions.concrete_actions.decoy_svchost import DecoySvchost
from csle_cyborg.shared.actions.concrete_actions.decoy_tomcat import DecoyTomcat
from csle_cyborg.shared.actions.concrete_actions.decoy_vsftpd import DecoyVsftpd
from csle_cyborg.shared.actions.concrete_actions.density_scout import DensityScout
from csle_cyborg.shared.actions.concrete_actions.escalate_action import EscalateAction
from csle_cyborg.shared.actions.concrete_actions.eternal_blue import EternalBlue
from csle_cyborg.shared.actions.concrete_actions.exploit_action import ExploitAction
from csle_cyborg.shared.actions.concrete_actions.ftp_directory_traversal import FTPDirectoryTraversal
from csle_cyborg.shared.actions.concrete_actions.haraka_rce import HarakaRCE
from csle_cyborg.shared.actions.concrete_actions.http_rfi import HTTPRFI
from csle_cyborg.shared.actions.concrete_actions.http_srfi import HTTPSRFI
from csle_cyborg.shared.actions.concrete_actions.juicy_potato import JuicyPotato
from csle_cyborg.shared.actions.concrete_actions.pingsweep import Pingsweep
from csle_cyborg.shared.actions.concrete_actions.port_scan import Portscan
from csle_cyborg.shared.actions.concrete_actions.remote_code_execution_on_smtp import RemoteCodeExecutionOnSMTP
from csle_cyborg.shared.actions.concrete_actions.restore_from_backup import RestoreFromBackup
from csle_cyborg.shared.actions.concrete_actions.sig_check import SigCheck
from csle_cyborg.shared.actions.concrete_actions.sql_injection import SQLInjection
from csle_cyborg.shared.actions.concrete_actions.ssh_brute_force import SSHBruteForce
from csle_cyborg.shared.actions.concrete_actions.stop_process import StopProcess
from csle_cyborg.shared.actions.concrete_actions.stop_service import StopService
from csle_cyborg.shared.actions.concrete_actions.v4l2_kernel_exploit import V4L2KernelExploit
from csle_cyborg.shared.actions.game_actions.create_agent import CreateAgent
from csle_cyborg.shared.actions.game_actions.game_action import GameAction
from csle_cyborg.shared.actions.game_actions.game_echo import GameEcho
from csle_cyborg.shared.actions.game_actions.game_sleep import GameSleep
from csle_cyborg.shared.actions.game_actions.get_true_state import GetTrueState
from csle_cyborg.shared.actions.game_actions.list_agents import ListAgents
from csle_cyborg.shared.actions.game_actions.list_all_sessions import ListAllSessions
from csle_cyborg.shared.actions.game_actions.reset_game import ResetGame
from csle_cyborg.shared.actions.global_actions.create_game import CreateGame
from csle_cyborg.shared.actions.global_actions.global_action import GlobalAction
from csle_cyborg.shared.actions.global_actions.global_echo import GlobalEcho
from csle_cyborg.shared.actions.global_actions.global_sleep import GlobalSleep
from csle_cyborg.shared.actions.global_actions.list_games import ListGames
from csle_cyborg.shared.actions.global_actions.shutdown_game import ShutdownGame
from csle_cyborg.shared.actions.green_actions.green_connection import GreenConnection
from csle_cyborg.shared.actions.green_actions.green_ping_sweep import GreenPingSweep
from csle_cyborg.shared.actions.green_actions.green_port_scan import GreenPortScan
from csle_cyborg.shared.actions.local_shell_actions.local_shell_sleep import LocalShellSleep
from csle_cyborg.shared.actions.local_shell_actions.local_shell_echo import LocalShellEcho
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.get_pid import GetPid
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.get_shell import GetShell
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.get_uid import GetUid
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.local_time import LocalTime
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_action import MeterpreterAction
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_ip_config import MeterpreterIPConfig
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_ps import MeterpreterPS
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.meterpreter_reboot import MeterpreterReboot
from csle_cyborg.shared.actions.msf_actions_folder.meterpreter_actions_folder.sys_info import SysInfo
from csle_cyborg.shared.actions.msf_actions_folder.msf_persistence_folder.service_persistence_windows import ServicePersistenceWindows
from csle_cyborg.shared.actions.msf_actions_folder.msf_privilege_escalation_folder.msf_privilege_escalation import MSFPrivilegeEscalation
from csle_cyborg.shared.actions.msf_actions_folder.msf_scanner_folder.msf_scanner import MSFScanner
from csle_cyborg.shared.actions.msf_actions_folder.msf_scanner_folder.msf_ping_sweep import MSFPingsweep
from csle_cyborg.shared.actions.msf_actions_folder.msf_scanner_folder.msf_port_scan import MSFPortscan
from csle_cyborg.shared.actions.msf_actions_folder.msf_scanner_folder.tomcat_credential_scanner import TomcatCredentialScanner
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.ms17_010_psexec import MS17_010_PSExec
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.msf_eternal_blue import MSFEternalBlue
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.ps_exec import PSExec
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.remote_code_execution import RemoteCodeExecution
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.ruby_on_rails import RubyOnRails
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.samba_usermap_script import SambaUsermapScript
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.ssh_login_exploit import SSHLoginExploit
from csle_cyborg.shared.actions.msf_actions_folder.remote_code_execution_folder.tomcat_exploit import TomcatExploit
from csle_cyborg.shared.actions.msf_actions_folder.msf_action import MSFAction
from csle_cyborg.shared.actions.msf_actions_folder.msf_auto_route import MSFAutoroute
from csle_cyborg.shared.actions.msf_actions_folder.upgrade_to_meterpreter import UpgradeToMeterpreter
from csle_cyborg.shared.actions.shell_actions_folder.delete_file_windows import DeleteFileWindows
from csle_cyborg.shared.actions.shell_actions_folder.delete_file_linux import DeleteFileLinux
from csle_cyborg.shared.actions.shell_actions_folder.find_flag import FindFlag
from csle_cyborg.shared.actions.shell_actions_folder.kill_process_linux import KillProcessLinux
from csle_cyborg.shared.actions.shell_actions_folder.kill_process_windows import KillProcessWindows
from csle_cyborg.shared.actions.shell_actions_folder.shell_action import ShellAction
from csle_cyborg.shared.actions.shell_actions_folder.shell_echo import ShellEcho
from csle_cyborg.shared.actions.shell_actions_folder.shell_ps import ShellPS
from csle_cyborg.shared.actions.shell_actions_folder.shell_sleep import ShellSleep
from csle_cyborg.shared.actions.shell_actions_folder.account_manipulation_folder.add_user_linux import AddUserLinux
from csle_cyborg.shared.actions.shell_actions_folder.account_manipulation_folder.account_manipulation import AccountManipulation
from csle_cyborg.shared.actions.shell_actions_folder.account_manipulation_folder.add_user_windows import AddUserWindows
from csle_cyborg.shared.actions.shell_actions_folder.account_manipulation_folder.disable_user_windows import DisableUserWindows
from csle_cyborg.shared.actions.shell_actions_folder.account_manipulation_folder.disable_user_linux import DisableUserLinux
from csle_cyborg.shared.actions.shell_actions_folder.account_manipulation_folder.remove_user_from_group_windows import RemoveUserFromGroupWindows
from csle_cyborg.shared.actions.shell_actions_folder.account_manipulation_folder.remove_user_from_group_linux import RemoveUserFromGroupLinux
from csle_cyborg.shared.actions.shell_actions_folder.credential_harvesting_folder.credential_harvesting import CredentialHarvesting
from csle_cyborg.shared.actions.shell_actions_folder.credential_harvesting_folder.read_passwd_file import ReadPasswdFile
from csle_cyborg.shared.actions.shell_actions_folder.credential_harvesting_folder.read_shadow_file import ReadShadowFile
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.internal_enumeration import InternalEnumeration
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.host_info_enumeration_folder.host_info_enumeration import HostInfoEnumeration
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.host_info_enumeration_folder.system_info import SystemInfo
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.host_info_enumeration_folder.uname import Uname
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.network_info_enumeration_folder.if_config import IFConfig
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.network_info_enumeration_folder.ip_config import IPConfig
from csle_cyborg.shared.actions.shell_actions_folder.internal_enumeration_folder.network_info_enumeration_folder.network_info_enumeration import NetworkInfoEnumeration
from csle_cyborg.shared.actions.shell_actions_folder.network_scan_folder.network_scan import NetworkScan
from csle_cyborg.shared.actions.shell_actions_folder.network_scan_folder.nmap_scan import NmapScan
from csle_cyborg.shared.actions.shell_actions_folder.network_scan_folder.ping_sweep import PingSweep
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.credential_access_folder.brute_force_access_folder.brute_force_access import BruteForceAccess
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.credential_access_folder.brute_force_access_folder.ssh_hydra_brute_force import SSHHydraBruteForce
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.credential_access_folder.credential_access import CredentialAccess
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.credential_access_folder.ssh_access import SSHAccess
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.netcat_connect import NetcatConnect
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.open_connection import OpenConnection
from csle_cyborg.shared.actions.shell_actions_folder.open_connection_folder.smb_anonymous_connection import SMBAnonymousConnection
from csle_cyborg.shared.actions.shell_actions_folder.persistence_folder.persistence import Persistence
from csle_cyborg.shared.actions.shell_actions_folder.persistence_folder.schtasks import Schtasks
from csle_cyborg.shared.actions.shell_actions_folder.service_manipulation_folder.service_manipulation import ServiceManipulation
from csle_cyborg.shared.actions.shell_actions_folder.service_manipulation_folder.shell_stop_service import ShellStopService
from csle_cyborg.shared.actions.shell_actions_folder.service_manipulation_folder.start_service import StartService
from csle_cyborg.shared.actions.shell_actions_folder.shell_privilege_escalation_folder.shell_privilege_escalation import ShellPrivilegeEscalation
from csle_cyborg.shared.actions.shell_actions_folder.shell_privilege_escalation_folder.dirty_cow_privilege_escalation import DirtyCowPrivilegeEscalation
from csle_cyborg.shared.actions.shell_actions_folder.shell_privilege_escalation_folder.linux_kernel_privilege_escalation import LinuxKernelPrivilegeEscalation
from csle_cyborg.shared.actions.shell_actions_folder.delete_file_linux import DeleteFileLinux
from csle_cyborg.shared.actions.shell_actions_folder.delete_file_windows import DeleteFileWindows
from csle_cyborg.shared.actions.shell_actions_folder.find_flag import FindFlag
from csle_cyborg.shared.actions.shell_actions_folder.kill_process_linux import KillProcessLinux
from csle_cyborg.shared.actions.shell_actions_folder.kill_process_windows import KillProcessWindows
from csle_cyborg.shared.actions.shell_actions_folder.shell_action import ShellAction
from csle_cyborg.shared.actions.shell_actions_folder.shell_echo import ShellEcho
from csle_cyborg.shared.actions.shell_actions_folder.shell_ps import ShellPS
from csle_cyborg.shared.actions.shell_actions_folder.shell_sleep import ShellSleep
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_alerts import GetAlerts
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_file_info import GetFileInfo
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_host_list import GetHostList
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_local_groups import GetLocalGroups
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_os_info import GetOSInfo
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_process_info import GetProcessInfo
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_process_list import GetProcessList
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_process_list_for_username import GetProcessListForUsername
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_user_info import GetUserInfo
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_user_ntds import GetUserNTDS
from csle_cyborg.shared.actions.velociraptor_actions_folder.get_users import GetUsers
from csle_cyborg.shared.actions.velociraptor_actions_folder.host_monitoring_action import HostMonitoringAction
from csle_cyborg.shared.actions.velociraptor_actions_folder.kill_process_name import KillProcessName
from csle_cyborg.shared.actions.velociraptor_actions_folder.kill_process_pid import KillProcessPID
from csle_cyborg.shared.actions.velociraptor_actions_folder.velociraptor_action import VelociraptorAction
from csle_cyborg.shared.actions.velociraptor_actions_folder.velociraptor_artifact_action import VelociraptorArtifactAction
from csle_cyborg.shared.actions.velociraptor_actions_folder.velociraptor_disable_user import VelociraptorDisableUser
from csle_cyborg.shared.actions.velociraptor_actions_folder.velociraptor_poll import VelociraptorPoll
from csle_cyborg.shared.actions.velociraptor_actions_folder.velociraptor_sleep import VelociraptorSleep
from csle_cyborg.shared.actions.action import Action
from csle_cyborg.shared.actions.session_action import SessionAction


import torch
from collections import deque
from csle_cyborg.shared.results import Results
from csle_cyborg.agents.wrappers.base_wrapper import BaseWrapper


class RewardShape(BaseWrapper):
    def __init__(self, env=None, agent=None):
        super().__init__(env, agent)
        self.action_buffer = deque(maxlen=2)
        self.observation_buffer = deque(maxlen=2)

    def step(self, agent=None, action=None) -> Results:
        result = self.env.step(agent, action)

        self.action_buffer.append(action)
        if torch.is_tensor(result.observation):
            self.observation_buffer.append(result.observation.tolist())
        else:
            self.observation_buffer.append(result.observation)

        if result.reward == 1.0:
            result.reward = 10.0
            result.done = True
            return result

        if len(self.action_buffer) == 2:
            if self.observation_buffer[0] == self.observation_buffer[1]:
                if self.action_buffer[0] == self.action_buffer[1]:
                    result.reward = -0.05
                else:
                    result.reward = -0.01
            else:
                result.reward = 0.01
                if action == 32 or action == 33 or action == 34:
                    result.reward = 1.0
        elif len(self.observation_buffer) == 2 and len(self.action_buffer) < 2:
            if self.observation_buffer[0] == self.observation_buffer[1]:
                result.reward = -0.01
            else:
                result.reward = 0.01
        else:
            pass

        return result

    def reset(self, agent=None):
        result = self.env.reset(agent)
        self.action_buffer = deque(maxlen=2)
        self.observation_buffer = deque(maxlen=2)
        if torch.is_tensor(result.observation):
            self.observation_buffer.append(result.observation.tolist())
        else:
            self.observation_buffer.append(result.observation)
        return result
